from odoo import models, fields, api, Command, _
from odoo.api import Environment
from odoo.tools import exception_to_unicode, ormcache, plaintext2html, parse_version, find_in_path
from odoo.exceptions import AccessDenied, UserError, ValidationError
from odoo.addons.loempia import util
from odoo.addons.loempia.util import chatter_log, search_for_modules
from odoo.addons.saas_worker.util import email_on_exception
from odoo.addons.saas_worker.const import USER, APPS_URL, BUYAPPS_HOST, TIMEOUT, WEBSITE
from odoo.addons.saas_worker.auth_util import SaasAuth
from odoo.tools.misc import split_every


import os
import shutil
import logging
import requests
import subprocess
from datetime import datetime, timedelta
from textwrap import dedent
from functools import partial
from urllib.parse import urlparse
from psycopg2.extensions import TransactionRollbackError


from odoo.addons.base.models.ir_module import Module as ir_module_module
LICENSES = ir_module_module.license.args['selection'][:]
SCAN_LIMIT = 2000
SCAN_BATCH_SIZE = 100
GIT_TIMEOUT = 60 * 5  # 5 minutes


_logger = logging.getLogger(__name__)

try:
    find_in_path("bzr")
    HAS_BZR = True
except IOError:
    HAS_BZR = False

_NO_DEFAULT = object()

class Series(models.Model):
    _name = 'loempia.series'
    _description = 'Apps Version'
    _order = 'sequence'
    _path_clean = ['icon_path', 'server_path']

    def _default_sequence(self):
        return self.search_count([])

    name = fields.Char('Name', size=9, required=True)
    sequence = fields.Integer('Sequence', default=_default_sequence, required=True)
    default_license = fields.Selection(LICENSES, 'Default license', default='Other proprietary', required=True)
    restrict_to_license = fields.Selection(LICENSES, 'Restrict to license', required=False)
    icon_path = fields.Char('Relative icon path', default='static/description/icon.png')
    manifest_file = fields.Char('Manifest File', default='__manifest__.py,__openerp__.py', required=True)
    server_path = fields.Char('Server Path', default='odoo', required=True)
    product_name = fields.Char('Product Name', default='Odoo', required=True)
    active = fields.Boolean(default=False, help='Allows you to control the publication of the series.'
        'Set to false by default as we want to first check the sanity of the series before making'
        'it available to the users.')
    enterprise_apps = fields.One2many('loempia.module', 'series_id',
                                      domain=lambda self: [
                                          ('application', '=', True),
                                          ('template_id.maintainer_id', '=', self.env.ref('base.user_admin').id),
                                      ],
                                      string='Enterprise Apps')

    @api.model_create_multi
    def create(self, vals_list):
        vals_list = [util.path_cleaner(vals, self._path_clean) for vals in vals_list]
        res = super(Series, self).create(vals_list)

        # auto. resequence all series
        series = self.search([]).sorted(key=lambda x: parse_version(x.name),reverse=True)
        for i, record in enumerate(series):
            record.write({'sequence': i})

        return res

    def write(self, values):
        values = util.path_cleaner(values, self._path_clean)
        return super(Series, self).write(values)

    @api.model
    def get_default(self):
        ids = self.search([], limit=1)
        if not ids:
            return False
        return ids[0]

    _sql_constraints = [
        ('uniq_sequence', 'UNIQUE(sequence) deferrable initially deferred', 'sequence must be unique'),
    ]


class Repo(models.Model):
    _name = 'loempia.repo'
    _description = 'Apps Repository'
    _rec_name = 'url'
    _inherit = ['mail.thread']
    _path_clean = ['addons_subdir']

    def _default_maintainer_id(self):
        return self.env.uid

    def _default_series_id(self):
        self.env["loempia.series"].get_default()

    priority = fields.Integer('Priority', default=1)
    series_id = fields.Many2one('loempia.series', default=_default_series_id, string='Series', required=True, tracking=True)
    url = fields.Char('Repository', size=128, required=True, tracking=True)    # TODO check url format?
    addons_subdir = fields.Char('Addons Subdirectory', default='', help='If addons are not at root of the repository', tracking=True, groups='loempia.group_manager')
    vcs = fields.Selection([('git', 'Git'), ('bzr', 'Bazaar')],
                           string='VCS', default='git', required=True, tracking=True)
    module_ids = fields.One2many('loempia.module', 'repo_id', string='Modules', readonly=True)
    auto_approve = fields.Boolean('Auto approve new modules', default=False, groups='loempia.group_manager', tracking=True)
    scan = fields.Boolean('Auto scan', tracking=True)   # FIXME, disallow users to check this manually...
    scan_log = fields.Html('Scan Log', default='', sanitize_style=True)
    path = fields.Char(compute='_path', string='Local Path', groups='loempia.group_manager')
    snap_path = fields.Char(compute='_snap_path', string='snapshot path', groups='loempia.group_manager')
    maintainer_id = fields.Many2one('res.users', default=_default_maintainer_id, string='Maintainer', tracking=True)
    is_maintainer = fields.Boolean(compute='_is_maintainer', default=True)
    is_server = fields.Boolean(compute='_is_server', store=True, groups='loempia.group_manager')
    is_odoo_enterprise = fields.Boolean(compute='_is_odoo_enterprise', store=True, groups='loempia.group_manager')
    state = fields.Selection([(s, s.title()) for s in 'draft active error dead'.split()],
                             string='Status', default='draft', readonly=True, required=True,
                             tracking=True)
    to_scan = fields.Boolean(default=False, string='To Scan', help='Marks the repository to be scanned in priority by the cron')

    _sql_constraints = [
        ('uniq_url', 'UNIQUE(url)', 'This url has already been registered'),    # will be recreated in init()
    ]

    def _log_repo(self, level, notify=True):
        self.ensure_one()

        def wrap(msg, *args, **kwargs):
            chatter_log(self, level, notify)(msg, *args, **kwargs)
            self.scan_log = plaintext2html(msg % args)
        return wrap

    def _check_url(self):
        self.ensure_one()
        warn_only = self.env['res.users'].has_group('loempia.group_manager')
        try:
            if ' ' in self.url:
                raise util.CheckURLError('The repository URL cannot contain spaces.')
            checking_method = getattr(self, '_check_url_%s' % (self.vcs,))
            if checking_method:
                checking_method()
        except util.CheckURLError as e:
            if warn_only:
                self._log_repo('warning')('Warning: %s', exception_to_unicode(e))
            else:
                raise ValidationError(exception_to_unicode(e)) from None
        return True

    def _check_url_bzr(self):
        self.ensure_one()
        # deprecated
        raise util.CheckURLError('Bazaar repositories are no longer supported.')

    def _check_url_git(self):
        self.ensure_one()
        # TODO only support github?
        if not self.url.endswith('#' + self.series_id.name):
            raise util.CheckURLError(
                ("The repository URL must contain a branch matching exactly the series name. "
                 "This repository is set for series {series}. Its url should then end with #{series}. "
                 "For example https://github.com/acme/odoo-addons.git#{series} "
                 "is valid for the {series} series.").format(series=self.series_id.name)
            )

    @api.model_create_multi
    def create(self, vals_list):
        for vals in vals_list:
            user = self.env['res.users'].browse(vals.get('maintainer_id', 0))
            if user.exists() and not user.oauth_uid:
                raise UserError(_('You do not have an Oauth identity. Please contact apps@odoo.com.'))

            vals['priority'] = max(vals.get('priority', 1), 1)

        vals_list = [util.path_cleaner(vals, self._path_clean) for vals in vals_list]
        repositories = super(Repo, self).create(vals_list)
        for repo in repositories:
            repo._check_url()
        return repositories

    def write(self, values):
        if 'priority' in values:
            values['priority'] = max(values['priority'], 1)

        # Automatically update series according to branch in url
        if 'url' in values:
            series_name = values['url'].split('#')[1]
            series_id = self.env['loempia.series'].search([('name', '=', series_name)], limit=1)
            if series_id:
                values['series_id'] = series_id.id
        values = util.path_cleaner(values, self._path_clean)
        result = super(Repo, self).write(values)
        if any(k in values for k in 'vcs url series_id'.split()):
            self._check_url()

        return result

    def unlink(self):
        self.ensure_one()
        paths = self.sudo().read(['path']) or []
        paths = [os.path.dirname(p['path'].rstrip('/')) for p in paths]

        # check user has the right to delete these repo
        self.check_access_rights('unlink')
        self.check_access_rule('unlink')

        # force direct unlink for modules to avoid errors with inter-dependencies
        self.sudo().mapped('module_ids').unlink()

        # then delete it as admin to avoid access errors on mail.thread deleting
        result = super(Repo, self.sudo()).unlink()

        # cleanup disk
        for p in paths:
            if os.path.exists(p):
                shutil.rmtree(p[:-5] if p.endswith('/repo') else p)

        return result

    def copy_data(self, default=None):
        default = dict(default or {})
        vals_list = super().copy_data(default=default)
        if 'url' not in default:
            for vals in vals_list:
                vals['url'] += " (copy)"
        return vals_list

    @api.depends('url')
    def _is_server(self):
        repos = ('ssh://git@github.com/odoo/odoo.git#',) + {
            # kept as legacy
            'chs': ('/Users/chs/devel/odoo/odoo/master#',),
            'Malorne': ('/Users/Malorne/Odoo/odoo#',),
            'mge': ('/home/mge-odoo/Dev/odoo#',),
            'odoo': ('/home/odoo/src/odoo/odoo#',),
        }.get(USER, ())
        for repo in self:
            repo.is_server = repo.url.startswith(repos)

    @api.depends('url')
    def _is_odoo_enterprise(self):
        repos = (
            'ssh://git@github.com/odoo/enterprise.git#',
            'ssh://git@github.com/odoo/design-themes.git#',
        ) + {
            # kept as legacy
            'chs': ('/Users/chs/devel/odoo/enterprise/master#',),
            'Malorne': ('/Users/Malorne/Odoo/enterprise#',),
            'mge': ('/home/mge-odoo/Dev/enterprise#',),
            'odoo': ('/home/odoo/src/odoo/enterprise#',),
        }.get(USER, ())
        for repo in self:
            repo.is_odoo_enterprise = repo.url.startswith(repos)

    def _server_zip_ignore_list(self, path):
        self.ensure_one()
        assert self.is_server

        ad_subdir = 'addons'
        ad_path = os.path.join(path, ad_subdir)
        return [
            os.path.join(ad_subdir, l)
            for l in os.listdir(ad_path)
            if l != 'base' and os.path.isdir(os.path.join(ad_path, l))
        ]

    @api.depends('maintainer_id')
    def _is_maintainer(self):
        is_manager = self.env['res.users'].has_group('loempia.group_manager')
        for repo in self:
            repo.is_maintainer = is_manager or repo.maintainer_id.id == self.env.uid

    @api.depends('url')
    def _path(self):
        # TODO: This function creates the repositories, it should be done elsewhere !
        root_dir = self.env['ir.config_parameter'].get_param('loempia.repo_cache', default='/tmp/repo_cache')
        root_dir = util.clean_path(root_dir)
        if not os.path.exists(root_dir):
            os.makedirs(root_dir)

        # ensure the repo_cache is a shared-repository to speed-up the pulling
        if HAS_BZR and not os.path.exists(os.path.join(root_dir, '.bzr')):
            subprocess.call(['bzr', 'init-repo', root_dir])

        for repo in self:
            path = False
            if repo.url:
                base = os.path.join(root_dir, repo._url_symlink(root_dir, repo.url))
                if not os.path.exists(base):
                    os.mkdir(base)
                path = os.path.join(base, 'repo')
            repo.path = path

    # clean the url for a human readable symlink
    def _url_symlink(self, root_dir, url):
        prefixes = ['lp:', 'https:']
        allowed_symbols = ['.', '-']
        for p in prefixes:
            if url.startswith(p):
                url = url[len(p):]
        url = "".join([c if c.isalnum() or c in allowed_symbols else '_' for c in url]).strip('_')
        return os.path.join(root_dir, url)

    @api.depends('path')
    def _snap_path(self):
        # TODO: Same here, it creates the folder !
        for repo in self:
            path = os.path.normpath(os.path.join(repo.path, '..', 'snapshot'))
            # ensure path exists
            if not os.path.exists(path):
                os.mkdir(path)
            repo.snap_path = path

    @api.onchange('url', 'vcs')
    def onchange_url(self):
        func = getattr(self, '_clean_url_%s' % (self.vcs,), None)
        if self.url and func:
            self.url = func(self.url)

    @api.model
    def _update_purchases(self):
        Purchases = self.env['loempia.module.purchase']

        public_purchases = Purchases.search([('state', 'not in', ['draft', 'cancel', 'exception']), ('user_id', '=', False)])
        order_ids_list = public_purchases.mapped('order_id')
        last_update_time = self.env['ir.config_parameter'].get_param('loempia.last_update_time', fields.Datetime.now())

        params = {'order_ids_list': order_ids_list, 'last_update_date': last_update_time}
        url = '{PROD}/openerp_apps/purchase_statistics'.format(PROD=BUYAPPS_HOST)
        req = requests.post(url, data=params, auth=SaasAuth(), timeout=TIMEOUT)
        purchases = req.json()

        # The below section is very sensitive to concurrent access errors,
        # with people creating new purchases all the time by adding modules
        # to their cart. Since the first section only contains reading
        # operations, we can safely commit here in order to shorten the time
        # required for the second section transaction.
        self.env.cr.commit()
        confirmed_orders = {}
        for (module_id_str, module_purchases) in purchases.items():
            module_id = self.env['loempia.module'].browse(int(module_id_str))
            if module_id.exists():
                for p in module_purchases:
                    # We are using the commercial version of the module (either lastest before 13 or the current)
                    commercial_module_id = module_id._get_commercial_series_module_id().id
                    loempia_purchases = Purchases.search([('order_id', '=', p['order_id']), ('module_id', '=', commercial_module_id)])
                    user_id = self.env['res.users'].search([('oauth_uid', '=', str(p['user_id']))], limit=1)  # Match the oauth user_id, which is a string

                    if p.get('user_login') and p.get('user_id') and not user_id:  # when user has bought without redirect to apps.
                        user_id = self.env['res.users'].search([('login', '=', p['user_login'])])  # Match the email since we can't create a new account with same login
                        if user_id:
                            user_id.oauth_uid = str(p['user_id'])  # Update the oauth_uid of the old account, apparently they use another one with this address now.
                        else:
                            user_id = self.env['res.users']._signup_create_user(
                                dict(
                                    login=p['user_login'],
                                    email=p['user_login'],
                                    oauth_uid=str(p['user_id']),
                                    name=p['user_login'],
                                    oauth_provider_id=self.env.ref('auth_oauth.provider_openerp').id
                                )
                            )

                    # If there was no purchase for this product in this order, we create one.
                    # Probably it was added manually on the sale.order, or it is a sale.order
                    # that dates back from before this commit went live.
                    if not loempia_purchases:
                        lp = Purchases.create({
                            'order_id': p['order_id'],
                            'order_name': p['order_name'],
                            'module_id': commercial_module_id,
                            'referrer_module_id': commercial_module_id,
                            'template_id': module_id.template_id.id,
                            'product_id': p['product_id'],
                            'date_order': p['date_order'],
                            'quantity': p['product_uom_qty'],
                            'price_unit': p['price_unit'],
                            'state': p['state'] if p['state'] != 'sale' else 'confirmed',
                            'user_id': user_id.id,
                        })
                        confirmed_orders[lp.order_id] = confirmed_orders.get(lp.order_id, []) + [lp]
                    else:
                        # Update the existing purchase
                        # There can only be one loempia purchase related to a given combination
                        # of order_id and product_id. Duplicates comes from the cart (draft state).
                        lp = loempia_purchases[0]
                        if lp.state != 'redeemed':
                            vals = {
                                'state': p['state'] if p['state'] != 'sale' else 'confirmed',
                                'order_name': p['order_name'],
                                'quantity': p['product_uom_qty'],
                            }
                            # With the new flow where we create products in the prod on the fly, we
                            # will create purchases of modules that do not have a product_id yet on the
                            # first purchase. This ensures that we set the product_id correctly on the
                            # loempia_module_purchase
                            if not lp.product_id:
                                vals['product_id'] = p['product_id']
                            if user_id:
                                vals['user_id'] = user_id.id
                            if lp.state == 'draft':
                                vals['date_order'] = fields.Datetime.now()
                                confirmed_orders[lp.order_id] = confirmed_orders.get(lp.order_id, []) + [lp]
                            lp.write(vals)

                        # Now that the purchase has been confirmed, we can clean the draft ones
                        loempia_purchases[1:].filtered(lambda p: p.state == 'draft').unlink()

                # Similarily to the module_purchase, we make sure that the produt_id is set to the module
                # when confirming the first purchase of the module.
                module_id.product_id = module_purchases[0]['product_id']
                module_id.template_id._update_previous_version(module_id, update_pricing=False, update_purchases=False)
        # update the last_update_time 2 hours-prior from now as a fallback if some orders are not properly synchronized.
        last_time_update = datetime.now() - timedelta(hours=2)
        self.env['ir.config_parameter'].set_param('loempia.last_update_time', fields.Datetime.to_string(last_time_update))

        return True

    @api.model
    def _send_stats_mailing(self):
        env = self.env
        most_downloaded_ids = search_for_modules(env, 'modules', limit=3, **dict(order='Downloads'))
        best_seller_ids = search_for_modules(env, 'modules', limit=3, **dict(order='Relevance'))
        best_seller_themes_ids = search_for_modules(env, 'themes', limit=3, **dict(order='Relevance'))
        newest_ids = search_for_modules(env, 'modules', limit=3, **dict(order='Newest'))
        newest_themes_ids = search_for_modules(env, 'themes', limit=3, **dict(order='Newest'))

        Module = self.env['loempia.module']

        mail_template = self.env.ref('loempia.apps_newsletter')
        # Cheat the context to pass multiple modules browserecords
        render_ctx = dict(
            self.env.context,
            most_downloaded_ids=Module.browse(most_downloaded_ids),
            best_seller_ids=Module.browse(best_seller_ids),
            best_seller_themes_ids=Module.browse(best_seller_themes_ids),
            newest_ids=Module.browse(newest_ids),
            newest_themes_ids=Module.browse(newest_themes_ids),
            community=True,
        )
        res_ids = Module.browse(best_seller_ids[0]).ids
        RenderMixin = self.env['mail.render.mixin'].with_context(**render_ctx)
        rendered_template = RenderMixin._render_template(
            mail_template.body_html,
            'loempia.module',
            res_ids
        )

        # Computing customers emails
        customer_ids = self.env['loempia.module.purchase'].read_group([], fields=['user_id'], groupby='user_id')
        customer_ids = [c['user_id'][0] for c in customer_ids if c['user_id']]
        customer_ids = self.env['res.users'].browse(customer_ids)

        # Look for customers already in the mailing list (to keep their opt-out)
        apps_mailing_list = self.env.ref('loempia.apps_mailing_list')
        Contacts = self.env['mailing.contact']
        already_registered_emails = Contacts.search_read([('list_ids', '=', apps_mailing_list.id)], ['email'])
        already_registered_emails = [c['email'] for c in already_registered_emails]

        new_contacts = [{'name': c.name, 'email': c.login, 'list_ids': [Command.set(apps_mailing_list.ids)]} for c in customer_ids if c.email not in already_registered_emails]
        for c in new_contacts:
            Contacts.create(c)

        self.env['mailing.mailing'].with_context(active_ids=res_ids[0]).create({
            'name': 'Odoo Apps Newsletter ' + datetime.today().strftime('%B %Y'),
            'subject': mail_template.subject,
            'email_from': 'noreply@odoo.com',
            'body_html': rendered_template.get(res_ids[0]),
            'state': 'in_queue',
            'contact_list_ids': [Command.set(apps_mailing_list.ids)],
            'schedule_type': 'now'
        })

    @api.model
    @email_on_exception()
    def _scan_all(self):
        self._accept_github_repository_invitations()
        # intended to be called as cron job
        domain = ['|', ('scan', '=', True), ('to_scan', '=', True), ('maintainer_id.banned', '=', False)]
        all_repos_to_scan = self.search(domain, order='to_scan desc,write_date', limit=SCAN_LIMIT)
        retry = self.browse()
        for repo_ids in split_every(SCAN_BATCH_SIZE, all_repos_to_scan, piece_maker=self.browse):
            for repo in repo_ids:
                try:
                    if not repo.to_scan:
                        repo._scan_repo(action='scan')
                    else:
                        repo._scan_repo(action='scan', force_scan=True, notify=False)
                except TransactionRollbackError:
                    _logger.exception('Failed to scan %r. Put in retry queue', repo)
                    retry |= repo

            if len(retry):
                _logger.info('Retry scan of %r', retry)
                for repo in retry:
                    try:
                        repo._scan_repo(action='scan')
                    except TransactionRollbackError:
                        _logger.exception('Re-Failed to scan %r. ignore', repo)
                        repo.write(dict(state='error'))
            # Commit the change to avoid CPU error
            self.env.cr.commit()

        return True

    def _accept_github_repository_invitations(self):
        # https://github.com/blog/2170-repository-invitations
        # https://developer.github.com/v3/repos/invitations/#list-a-users-repository-invitations
        token = self.env['ir.config_parameter'].sudo().get_param('apps.invitations_token')
        headers = {'Accept': 'application/vnd.github.swamp-thing-preview+json', 'Authorization': 'token %s' % token}
        invits = requests.get('https://api.github.com/user/repository_invitations', headers=headers)
        if invits.status_code == 200:
            for invit in invits.json():
                accept = requests.patch(invit['url'], headers=headers)
                if accept.status_code == 204:
                    msg = "Invitation for repo %s from %s (%s) has been accepted." % (invit['repository']['full_name'], invit['inviter']['login'], invit['created_at'])
                else:
                    msg = "Invitation for repo %s has failed: %s - %s" % (invit['repository']['full_name'], accept.status_code, accept.content)
                _logger.info(msg)
        else:
            msg = "Cannot list github invitations for github: %s - %s" % (invits.status_code, invits.content)
            _logger.info(msg)

    def scan_repo(self):
        for repo in self:
            repo._scan_repo(action='scan', force_scan=True, notify=False)
        return True

    def _scan_repo(self, action, force_scan=False, exception_state='error', set_auto_scan=False, notify=True):
        self.ensure_one()
        error_data = dict(state=exception_state)
        repo = None

        with self.pool.cursor() as repo_cr:
            error = _logger.error
            try:
                is_manager = self.env['res.users'].has_group('loempia.group_manager')
                repo_env = Environment(repo_cr, self.env.uid, self.env.context or {})
                repo = self.with_env(repo_env).sudo()

                _logger.info('update %s repo %s', repo.vcs, repo.url)

                if repo.to_scan:
                    repo.to_scan = False
                    repo_cr.commit()

                Modules = repo_env['loempia.module']
                # Guard the repo update with an ExclusiveWriteLock on all modules to make sure we will never
                # be blocked by a concurrent download count update. Lock will be released at commit.
                repo_cr.execute("SELECT id FROM %s WHERE id = %%s FOR UPDATE" % (self._table,), (self.id,))
                repo_cr.execute("SELECT id FROM %s WHERE repo_id = %%s FOR UPDATE" % (Modules._table,), (self.id,))

                if repo.state == 'error':
                    # deactivate auto scan if an error occurs twice in a row
                    error_data['scan'] = False

                warn = repo._log_repo('warning', notify=notify)
                error = repo._log_repo('error', notify=notify)

                if force_scan and repo.vcs == 'git':  # not via cron
                    self._accept_github_repository_invitations()

                func = getattr(repo, '_update_{0.vcs}'.format(repo))
                if repo.is_server and not is_manager:
                    repo.write(error_data)
                    error('You cannot register a repository containing an Odoo server, only modules are allowed.')
                    return False

                new_revs = []
                try:
                    new_revs = func()
                    if not new_revs and not force_scan:
                        _logger.info('repo %s already up to date', repo.url)
                        # RAR: might need a rework in 13.0 - new field `scan_date`?
                        repo.scan = repo.scan
                        return True
                except util.VCSError as e:
                    repo.write(error_data)
                    error(exception_to_unicode(e))
                    return False

                warnings = []
                update = partial(Modules.sudo().create_from_path, repo.id)

                path = os.path.join(repo.path, repo.addons_subdir or '')
                if force_scan:
                    # during a force_scan, we pass through the whole repo, regardless of the modifications
                    # However, since we are already sync'ed with the remote, a deleted module will not appear on disk
                    # and we need to make use of new_revs
                    new_revs = list(set(os.listdir(path) + new_revs))
                dir_paths = [os.path.join(path, d) for d in new_revs]

                if repo.is_server:
                    Modules._create_server_module(repo)
                    # Since 8.0, we have test modules in the subrepo {server_path}/addons (server_path in (openerp,odoo))
                    # In order to allows 3rd party devs to rely on those to add general tests to their own modules, we manually add
                    # the subrepo in the list of dirs that have to be scanned.
                    server_path = os.path.join(repo.path, repo.series_id.server_path, repo.addons_subdir)
                    subdir_list = os.listdir(server_path)
                    # We remove 'base' from the list as we create it with  _create_server_module
                    if 'base' in subdir_list:
                        subdir_list.remove('base')
                    dir_paths += [os.path.join(server_path, d) for d in subdir_list]

                def module_by_name(repo, module_name):
                    return repo.module_ids.filtered(lambda m: m.name == module_name)

                dirs = []
                for d in dir_paths:
                    # Unpublish the module automatically if it was already created and was then deleted from the repository.
                    if not os.path.isdir(d):
                        module_name = os.path.split(d.rstrip('/'))[-1]
                        module_by_name(repo, module_name).is_deleted = True
                    else:
                        for manifest_name in repo.series_id.manifest_file.split(','):
                            manifest_path = os.path.join(d, manifest_name.strip())
                            if os.path.isfile(manifest_path):
                                dirs.append(d)
                                break

                if not dirs and new_revs:
                    error("No module found in repository %s.\n"
                          "Each module needs to be in a separate folder at the root of the repository, "
                          "even if there is only one module.\n"
                          "If you have all the files of your module at the root of your repository, "
                          "you need to create a new folder with the name of your module and move all the files in it.",
                          repo.url)
                    return False

                invalid_mod_names = []
                repush = None
                while dirs:
                    d = dirs.pop(0)
                    try:
                        update_list = [directory.split('/')[-1] for directory in dirs]
                        update(d, update_list)
                        repush = None
                    except util.DependencyError as e:
                        # error due to unmet dependency, process later
                        if repush == d:
                            # ignore module
                            warnings.append(exception_to_unicode(e))
                            repush = None
                            continue
                        if repush is None:
                            repush = d
                        dirs.append(d)
                    except util.ModuleRepoError as e:
                        # module belongs to another repo, ignore
                        warnings.append(exception_to_unicode(e))
                        if repush == d:
                            repush = None
                    except util.ManifestError as e:
                        # ignore the module...
                        mod_name = os.path.basename(d)
                        invalid_mod_names.append(mod_name)
                        warnings.append('%s: %s' % (mod_name, exception_to_unicode(e)))
                        if repush == d:
                            repush = None

                if warnings:
                    warn('\n'.join(warnings))

                repo._invalidate_modules(invalid_mod_names)

                vals = dict(state='active')
                if set_auto_scan or repo.state == 'error':
                    vals['scan'] = True
                if not warnings:
                    vals['scan_log'] = None
                repo.write(vals)

                # automatically hide our own modules from the store
                if repo.is_server or repo.is_odoo_enterprise:
                    repo.mapped('module_ids').write({'status': 'blocked', 'is_published': False})

            except Exception as e:
                error("Failed to %s repository: %s", action, exception_to_unicode(e), exc_info=True)
                if repo:
                    repo.write(error_data)
                    repo.module_ids.write({'is_published': False})
                return False
        return True

    def _invalidate_modules(self, invalid_mod_names):
        """
        Mark the given modules of a given repository as invalid, if permitted by their current state.

        params: list invalid_mod_names: the names of the modules to mark as invalid
        """
        self.ensure_one()
        if invalid_mod_names:
            invalid_modules = self.module_ids.filtered(lambda m: m.name in invalid_mod_names and m.status not in ['blocked', 'banned'])
            invalid_modules.write({'status': 'invalid'})

    @api.model
    def _validate_all(self):
        # intended to be called as cron job
        domain = [('state', '=', 'draft'), ('scan', '!=', False)]
        all_draft_repos = self.search(domain, order='write_date', limit=SCAN_LIMIT)
        for repos in split_every(SCAN_BATCH_SIZE, all_draft_repos, piece_maker=self.browse):
            for repo in repos:
                repo.validate()
            self.env.cr.commit()
        return True

    def validate(self):
        self.ensure_one()
        is_funtional = self.env['res.users'].has_group('loempia.group_functional')
        if not is_funtional:
            raise AccessDenied()
        if not self.scan:
            self._scan_repo(action='validate',
                            force_scan=True,
                            set_auto_scan=True)
        return True

    def clean_tree(self):
        self.ensure_one()
        is_manager = self.env['res.users'].has_group('loempia.group_manager')
        if not is_manager:
            raise AccessDenied()
        if os.path.exists(self.path):
            shutil.rmtree(self.path)
        return True

    def rip(self):
        self.write({'state': 'dead', 'scan': False})

    def _update_bzr(self):
        repo = self
        assert repo.vcs == 'bzr'

        def run(cmd, cwd):
            try:
                _logger.debug('running %r', cmd)
                subprocess.check_output(cmd, cwd=cwd, stderr=subprocess.STDOUT)
            except subprocess.CalledProcessError as cpe:
                # from pipes import quote  # as shell_quote_argument
                # cmd = ' '.join(quote(x) for x in cmd)
                scan_error = dedent("""Error when updating bzr from {repo.url}
                                returncode: {e.returncode}
                                output:
                                {e.output}
                                Consult our FAQ at {url}/apps/faq .
                """).strip().format(e=cpe, repo=repo, url=APPS_URL)
                raise util.VCSError(scan_error)

        if not os.path.exists(repo.path):
            path, dir_ = os.path.split(repo.path.rstrip('/'))
            cmd = 'bzr', 'branch', '-q', repo.url, dir_
            try:
                run(cmd, path)
                return list(os.listdir(repo.path))
            except Exception:
                if os.path.exists(repo.path):
                    shutil.rmtree(repo.path)
                raise
        else:
            # TODO bzr missing is slow. Use lp api (with oauth) and compare revid
            # requests.get("https://api.launchpad.net/devel/branches?ws.op=getByUrl&url=lp:{repo.url}".format(repo=repo)).json()['last_scanned_id']
            # open(os.path.join(repo.path, '.bzr', 'branch', 'last-revision').read().strip().split()[-1]
            cmd = 'bzr', 'missing', '-q', '--theirs-only', '--line', repo.url
            with open(os.devnull, 'r') as dn:
                rc = subprocess.call(cmd, cwd=repo.path, stdout=dn, stderr=subprocess.STDOUT)

            if rc == 0:
                # no new revisions available
                return []

            cmd = 'bzr', 'pull', '-q', '--overwrite', repo.url
            run(cmd, cwd=repo.path)
            return list(os.listdir(repo.path))

    def _clean_url_bzr(self, url):
        url = (url or '').rstrip('/')
        prefixes = [
            'https://code.launchpad.net/',
            'bzr+ssh://bazaar.launchpad.net/',
        ]
        for prefix in prefixes:
            if url.startswith(prefix):
                return 'lp:' + url[len(prefix):]
        return url

    def _update_git(self):
        repo = self
        assert repo.vcs == 'git'

        def run(cmd, cwd, shell=False):
            try:
                _logger.debug('running %r', cmd)
                return subprocess.check_output(cmd, cwd=cwd, stderr=subprocess.STDOUT, shell=shell, timeout=GIT_TIMEOUT)
            except subprocess.CalledProcessError as cpe:
                scan_error = dedent("""Error when updating git from {repo.url}
                                        returncode: {e.returncode}
                                        output:
                                        {e.output}
                                        Consult our FAQ at {url}/apps/faq .
                """).strip().format(e=cpe, repo=repo, url=APPS_URL)
                raise util.VCSError(scan_error)

        url, _, branch = repo.url.rpartition('#')
        if not os.path.exists(repo.path):
            path, dir_ = os.path.split(repo.path.rstrip('/'))
            cmd = 'git', '-c', 'core.askpass=/bin/true', 'clone', '--quiet', '--branch', branch, '--single-branch', url, dir_
            try:
                run(cmd, path)
            except Exception:
                if os.path.exists(repo.path):
                    shutil.rmtree(repo.path)
                raise
            return list(os.listdir(repo.path))
        else:
            cmd = 'git', '-c', 'core.askpass=/bin/true', 'fetch', '--quiet', '--force', '--no-tags'     # no purge !
            run(cmd, repo.path)

            cmd = "git diff --name-only HEAD origin/%s | awk -F '/' '{print $1}' | uniq" % branch
            rc = run(cmd, cwd=repo.path, shell=True)
            rc = rc.decode().split('\n')[:-1]
            if rc:
                # new revisions available
                cmd = 'git', 'reset', '--quiet', '--hard', 'origin/' + branch
                run(cmd, repo.path)

            return rc

    def _clean_url_git(self, url):
        url = (url or '').rstrip('/')
        prefixes = {
            # 'git@github.com:': 'https://github.com/',
        }
        for k, v in prefixes.items():
            if url.startswith(k):
                return v + url[len(k):]
        # In order to stop people from disclosing their git credentials,
        # replace the form http(s)[credentials]@ with simple http(s)
        parsed = urlparse(url)
        if parsed.scheme != 'ssh':
            return parsed._replace(netloc=parsed.hostname).geturl()
        return url

    def _publish_repo(self):
        self.module_ids.write({'is_published': True})

    def _unpublish_repo(self):
        self.module_ids.write({'is_published': False})
